@@ -1,7 +1,7 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
 ;; File:          dired-shell.el
-;; Dired Version: #Revision: 7.9 $
+;; Dired Version: 7.15
 ;; RCS:
 ;; Description:   Commands for running shell commands on marked files.
 ;;
@@ -20,7 +20,9 @@
 	(require 'ps-print)
 	(concat ps-lpr-command
 		" "
-		(ps-flatten-list (mapcar 'ps-eval-switch ps-lpr-switches))))
+		(mapconcat 'identity
+			   (ps-flatten-list (mapcar 'ps-eval-switch ps-lpr-switches))
+			   " ")))
     (error
      (concat
       (if (boundp 'lpr-command)
@@ -59,13 +61,17 @@
 This is used by `dired-do-print' to determine the default print command for
 printing the marked files.")
 
-(defvar dired-auto-shell-command-alist nil
+(defcustom dired-auto-shell-command-alist nil
   "*Alist of regexps and command lists to guess shell commands.
 Each element of this list should be a list of regular expression, and a list
 of guesses for shell commands to be used if the file name matches the regular
 expression. The list of guesses is evalled. This alist is appended to the front
 of dired-default-auto-shell-command-alist before prompting for each shell
-command.")
+command."
+  :group 'dired
+  :type '(repeat (cons (regexp)
+		       (choice (repeat (string))
+			       (sexp)))))
 
 (defvar dired-default-auto-shell-command-alist
   (list
@@ -110,7 +116,10 @@ command.")
    '("\\.gp$" "gnuplot")
    '("\\.gif$" "xv")			; view gif pictures
    '("\\.fig$" "xfig")			; edit fig pictures
+   '("\\.ps\\(?:\.gz\\|\.Z\\)?$" "gv")	; gv reads gzipped files.
    '("\\.ps$" "ghostview")
+   '("\\.pdf$" "acroread")		; Adobe acrobat (pdftex)
+   '("\\.lect$" "LecternClient")	; DEC SRC Virtual Paper
 
    ;; Typesetting.  For printing documents, see dired-print-program-alist.
    '("\\.tex$" "latex" "tex")
@@ -149,7 +158,8 @@ Set this to nil to turn off shell command guessing.")
   "Wildcard character used by dired shell commands.
 Indicates where file names should be inserted.")
 
-(defvar dired-shell-command-separators '(?\  ?| ?> ?< ?& ?;)
+(defvar dired-shell-command-separators '(?\  ?| ?> ?< ?& ?;
+					     )
   "Defines the start of a string specifying a word in a shell command.")
 
 (defvar dired-trans-map
@@ -183,16 +193,18 @@ If nil, such files are not marked.")
 ;;; Utility functions and Macros
 
 (defun dired-shell-quote (filename)
-  ;; Quote a file name for inferior shell (see variable shell-file-name).
-  ;; Quote everything except POSIX filename characters.
-  ;; This should be safe enough even for really weird shells.
-  (let ((result "") (start 0) end)
-    (while (string-match "[^---0-9a-zA-Z_./]" filename start)
-      (setq end (match-beginning 0)
-	    result (concat result (substring filename start end)
-			   "\\" (substring filename end (1+ end)))
-	    start (1+ end)))
-    (concat result (substring filename start))))
+  (if (fboundp 'shell-quote-argument)
+      (shell-quote-argument filename)
+    ;; Quote a file name for inferior shell (see variable shell-file-name).
+    ;; Quote everything except POSIX filename characters.
+    ;; This should be safe enough even for really weird shells.
+    (let ((result "") (start 0) end)
+      (while (string-match "[^---0-9a-zA-Z_.+/]" filename start)
+	(setq end (match-beginning 0)
+	      result (concat result (substring filename start end)
+			     "\\" (substring filename end (1+ end)))
+	      start (1+ end)))
+      (concat result (substring filename start)))))
 
 (defun dired-uniquefy-list (list)
   ;; Returns list, after removing 2nd and higher occurrences
@@ -514,9 +526,10 @@ This is particularly usefull if we are sending file names to a remote shell."
 	    (progn
 	      (setq mode-line-process '(": %s"))
 	      (start-process "Shell" (current-buffer)
-			     shell-file-name "-c" command))
-	  (call-process shell-file-name nil t nil "-c" command))))))
+			     shell-file-name shell-command-switch command))
+	  (call-process shell-file-name nil t nil shell-command-switch command))))))
 
+;;;###autoload
 (defun dired-run-shell-command (command dir in-background &optional append)
   ;; COMMAND is shell command
   ;; DIR is directory in which to do the shell command.
@@ -551,7 +564,7 @@ This is particularly usefull if we are sending file names to a remote shell."
 	  (if append
 	      (progn
 		(goto-char (point-max))
-		(or (= (char-before (point)) ?\n) (bobp) (insert "\n")))
+		(or (equal (char-before (point)) ?\n) (bobp) (insert "\n")))
 	    (erase-buffer)
 	    (if (stringp in-background)
 		(progn
@@ -570,13 +583,15 @@ This is particularly usefull if we are sending file names to a remote shell."
 	(if append
 	    (progn
 	      (goto-char (point-max))
-	      (or (= (char-before (point)) ?\n) (bobp) (insert "\n")))
+	      (or (equal (char-before (point)) ?\n) (bobp) (insert "\n")))
 	  (erase-buffer))
 	(dired-shell-call-process command dir)))))
 
 ;;; User commands
 
+;;;###autoload
 (defun dired-do-shell-command (command arg files &optional in-background)
+  ;; ARG = (64) means operate on each file, in (dired-dwim-target-directory)
   ;; ARG = (16) means operate on each file, in its own directory.
   ;; ARG = (4) means operate on each file, but in the current
   ;;       default-directory.
@@ -595,15 +610,20 @@ will be sent.
 With a prefix of \\[universal-argument], a separate command for each file will
 be executed.
 
-With a prefix of \\[universal-argument] \\[universal-argument], a separate command will be sent for each file,
-and the command will be executed in the directory of that file.  The explicit
-command will be of the form 
+With a prefix of \\[universal-argument] \\[universal-argument], a
+separate command will be sent for each file, and the command will be
+executed in the directory of that file.  The explicit command will be
+of the form
 
                       cd dir; cmd file
 
 When prompting for the shell command, dired will always indicate the directory
 in which the command will be executed.
 
+With a prefix of \\[universal-argument] \\[universal-argument]
+\\[universal-argument], a eparate command will be sent for each file,
+and the command will be executed in the dwim directory.
+
 The following documentation depends on the settings of `dired-file-wildcard',
 `dired-shell-command-separators', `dired-trans-map', `dired-shell-prefix',
 `dired-shell-separator', and `dired-shell-postfix'. See the documentation for
@@ -657,16 +677,21 @@ The prefix character for file name transformers is always the same as
 
   (interactive
    (let ((on-each (or (equal '(4) current-prefix-arg)
-		      (equal '(16) current-prefix-arg)))
+		      (equal '(16) current-prefix-arg)
+		      (equal '(64) current-prefix-arg)))
 	 (files (dired-get-marked-files
 		 nil (and (not (consp current-prefix-arg))
 			  current-prefix-arg)))
-	 (dir (and (not (equal current-prefix-arg '(16)))
-		   (dired-current-directory))))
+	 (dir (or (and (equal current-prefix-arg '(64))
+		       (dired-dwim-target-directory))
+		  (and (not (equal current-prefix-arg '(16)))
+		       (dired-current-directory)))))
      (list
       (dired-read-shell-command
        (concat (if dir
-		   (format "! in %s " (dired-abbreviate-file-name dir))
+		   (format "! in %s " (replace-in-string
+                                       (dired-abbreviate-file-name dir)
+                                       "%" "%%"))
 		 "cd <dir>; ! ")
 	       "on "
 	       (if on-each "each ")
@@ -687,9 +712,11 @@ The prefix character for file name transformers is always the same as
 
   ;; Look out for remote file names.
   
-  (let* ((on-each (or (equal arg '(4)) (and (equal arg '(16)) 'dir)))
+  (let* ((on-each (or (equal arg '(4)) (and (equal arg '(16)) 'dir)
+		      (equal arg '(64))))
 	 (ufiles (mapcar 'dired-shell-unhandle-file-name files))
-	 (dir (dired-current-directory))
+	 (dir (or (and (equal arg '(64)) (dired-dwim-target-directory))
+		  (dired-current-directory)))
 	 (udir (dired-shell-unhandle-file-name dir)))
 
     (save-excursion ; in case `shell-command' changes buffer
@@ -708,7 +735,7 @@ The prefix character for file name transformers is always the same as
 	;; A better solution needs to be found.
 	(dired-run-shell-command
 	 (dired-shell-stuff-it command ufiles udir on-each)
-	 dir (if (equal arg '(16))
+	 dir (if (or (equal arg '(16)) (equal arg '(64)))
 		 (concat "cd <dir>; \"" command "\"\n\n")
 	       (concat "\"" command "\" in " dir "\n\n"))))
        (on-each
@@ -770,6 +797,7 @@ The prefix character for file name transformers is always the same as
 	(let ((dired-no-confirm '(revert-subdirs)))
 	  (dired-verify-modtimes)))))
 
+;;;###autoload
 (defun dired-do-background-shell-command (command arg files)
   "Like \\[dired-do-shell-command], but starts command in background.
 Note that you can type input to the command in its buffer.
@@ -806,6 +834,7 @@ This requires background.el from the comint package to work."
 
 ;;; Printing files
 
+;;;###autoload
 (defun dired-do-print (&optional arg command files)
   "Print the marked (or next ARG) files.
 Uses the shell command coming from variable `dired-print-program-alist'."
