@@ -1,7 +1,7 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
 ;; File:          dired-xemacs.el
-;; Dired Version: #Revision: 7.9 $
+;; Dired Version: 7.15
 ;; RCS:
 ;; Description:   dired functions for XEmacs
 ;; Author:        Mike Sperber <sperber@informatik.uni-tuebingen.de>
@@ -162,7 +162,7 @@ This shouldn't match socket or symbolic link lines (which aren't editable).")
      ["Compress/Uncompress" dired-do-compress t]
      ["Uuencode/Uudecode" dired-do-uucode t]
      ["Print..." dired-do-print t]
-     ["Change Mode..." dired-do-interactive-chmod t]
+     ["Change Mode..." dired-do-chmod t]
      ["Change Owner..." dired-do-chown t]
      ["Change Group..." dired-do-chgrp t]
      ["Load" dired-do-load t]
@@ -275,17 +275,21 @@ This shouldn't match socket or symbolic link lines (which aren't editable).")
   "Installs the Dired menu at the menubar."
   (if (null dired-help-menu)
       (dired-setup-menus))
-  (when (and (featurep 'menubar) current-menubar)
-    (mapcar
-     (function
-      (lambda (pair)
-	(let ((name (car pair))
-	      (menu (symbol-value (cdr pair))))
-	  (easy-menu-add (cons name menu)))))
-     dired-menubar-menus)
-    (unless (featurep 'infodock)
-      (add-menu-button '("Help") (list "---"))
-      (add-submenu '("Help") (cons "Dired" dired-help-menu)))))
+  (if (and (featurep 'menubar) current-menubar)
+      (progn
+	(let ((buffer-menubar (copy-sequence current-menubar)))
+	  (set-buffer-menubar buffer-menubar)
+	  (mapcar
+	   (function
+	    (lambda (pair)
+	      (let ((name (car pair))
+		    (menu (symbol-value (cdr pair))))
+		(add-submenu nil (cons name menu)))))
+	   dired-menubar-menus))
+ 	(unless
+ 	    (car (find-menu-item current-menubar '("Help" "Dired")))
+ 	  (add-menu-button '("Help") "--:shadowEtchedIn")
+ 	  (add-submenu '("Help") (cons "Dired" dired-help-menu))))))
 
 (add-hook 'dired-mode-hook 'dired-install-menubar)
 
@@ -434,22 +438,68 @@ If the file name is already flag, this unflags it."
 ;; Make this defined everywhere in the dired buffer.
 (define-key dired-mode-map '(meta button3) 'dired-mouse-get-target)
 
+
+(define-key dired-mode-map [(home)] (function (lambda ()
+				      (interactive)
+				      (beginning-of-line)
+				      (when dired-do-interactive-permissions
+					(dired-move-to-permissions ?u 1)))))
+
+(define-key dired-mode-map [(end)] (function (lambda ()
+				     (interactive)
+				     (beginning-of-line)
+				     (dired-move-to-filename))))
+
 ;;; Extent managment
 
+(defun dired-maybe-filename-start (start end)
+  "Return start of filename if indicated by 'dired-file-name extent.
+Presumably, this extent was inserted by `insert-directory'."
+  (let ((extents (dired-filename-extents start end)))
+    (and extents
+	 (extent-start-position (car extents)))))
+
+(defun dired-maybe-filename-end (start end)
+  "Return end of filename if indicated by 'dired-file-name extent.
+Presumably, this extent was inserted by `insert-directory'."
+  (let ((extents (dired-filename-extents start end)))
+    (and extents
+	 (extent-end-position (car extents)))))
+
+(defun dired-filename-extents (start end)
+  "Returns extents with DIRED-FILE-NAME property.
+START and END say where."
+  (let ((extents '()))
+    (map-extents #'(lambda (extent _)
+		     (if (extent-property extent 'dired-file-name)
+			 (setq extents (cons extent extents)))
+		     nil)
+		 nil
+		 start end)
+    extents))
+
+(defun dired-set-extent-text-properties (extent face)
+  (set-extent-face extent face)
+  (set-extent-property extent 'keymap dired-filename-local-map)
+  (set-extent-property extent 'highlight t)
+  (set-extent-property
+   extent 'help-echo
+   (concat
+    "button2 finds, button3 visits, "
+    "C-button2 file ops, [C-]shift-button1 marks/flags.")))
+
 (defun dired-set-text-properties (start end &optional face)
-  (let ((filename-extent (make-extent start end)))
-    (set-extent-face filename-extent (or face 'default))
-    (set-extent-property filename-extent 'dired-file-name t)
-    (set-extent-property filename-extent 'start-open t)
-    (set-extent-property filename-extent 'end-open t)
-    (set-extent-property filename-extent 'keymap dired-filename-local-map)
-    (set-extent-property filename-extent 'highlight t)
-    (set-extent-property
-     filename-extent 'help-echo
-     (concat
-      "button2 finds, button3 visits, "
-      "C-button2 file ops, [C-]shift-button1 marks/flags."))
-    filename-extent))
+  (let ((extent
+	 (let ((extents (dired-filename-extents start end)))
+	   (if extents
+	       (car extents)
+	     (let ((extent (make-extent start end)))
+	       (set-extent-property extent 'dired-file-name t)
+	       (set-extent-property extent 'start-open t)
+	       (set-extent-property extent 'end-open t)
+	       extent)))))
+    (dired-set-extent-text-properties extent (or face 'default))
+    extent))
 
 (defun dired-insert-set-properties (beg end)
   ;; Sets the extents for the file names and their properties
@@ -497,10 +547,9 @@ If the file name is already flag, this unflags it."
   (map-extents
    (function
     (lambda (extent maparg)
-      (if (extent-property extent 'dired-file-name)
-	  (delete-extent extent))
+      (delete-extent extent)
       nil))
-   nil start end))
+   nil start end nil nil 'dired-file-name))
 
 (defun dired-highlight-filename-mark (extent)
   (let ((mark
@@ -578,6 +627,23 @@ If the file name is already flag, this unflags it."
 	(goto-char (extent-end-position filename-extent))
       (and (null no-error) (error "No file on this line")))))
 
+(defun dired-move-to-permissions (domain direction)
+  (skip-chars-backward "^\r\n")
+  (let (extent)
+    (while (not (setq extent (map-extents (function
+					   (lambda (e p)
+					    (and
+					     (char-equal domain (extent-property e p))
+					     e)))
+					  (current-buffer)
+					  (point)
+					  (save-excursion
+					    (skip-chars-forward "^\r\n")
+					    (point))
+					  'dired-permissions)))
+      (forward-line direction))
+    (goto-char (extent-start-position extent))))
+
 ;;; Interactive chmod
 ;;; (based on ideas from Russell Ritchie's dired-chmod.el)
 
@@ -697,8 +763,48 @@ If the file name is already flag, this unflags it."
 
       (define-key keymap 'button2 mouse-toggle-name)
       (define-key keymap 'button3 mouse-menu-name)
+      (define-key keymap [(control ?p)] (function (lambda ()
+					  (interactive)
+					  (dired-permissions-arrow-move -1))))
+      (define-key keymap [(up)] (function (lambda ()
+				  (interactive)
+				  (dired-permissions-arrow-move -1))))
+      (define-key keymap [(control ?n)] (function (lambda ()
+					  (interactive)
+					  (dired-permissions-arrow-move 1))))
+      (define-key keymap [(down)] (function (lambda ()
+				    (interactive)
+				    (dired-permissions-arrow-move 1))))
+      (set-keymap-name keymap (char-to-string domain))
       keymap)))
-    
+
+(defun dired-permissions-arrow-move (direction)
+  (let ((domain (extent-property
+		 (extent-at (point) nil 'dired-permissions)
+		 'dired-permissions))
+	(here (point)))
+    (forward-line direction)
+    ;; /etc/rcN.d is ".", "..", and the rest all symlinks.
+    (if (catch 'found-one
+	  (while t
+	    (cond
+	      ;; Cannot modify symlink permissions.
+	      ((looking-at dired-re-sym) (forward-line direction))
+	      ((looking-at "^. [ \t0-9]*[^l][r-][w-]") (throw 'found-one t))
+	      (t (throw 'found-one nil)))))
+	(dired-move-to-permissions domain direction)
+      (goto-char here)
+      (if (save-excursion
+	    (forward-line direction)
+	    (looking-at dired-re-sym))
+	  (progn
+	    (forward-line direction)
+	    (dired-move-to-filename))
+	(dired-move-to-permissions domain direction)
+	(when signal-error-on-buffer-boundary
+	  ;; it'll do for now...
+	  (ding nil 'buffer-bound))))))
+
 (defvar dired-u-r-keymap nil "internal keymap for dired")
 (defvar dired-u-w-keymap nil "internal keymap for dired")
 (defvar dired-u-x-keymap nil "internal keymap for dired")
@@ -745,10 +851,203 @@ If the file name is already flag, this unflags it."
   (set-extent-face extent 'dired-face-permissions)
   (set-extent-property extent 'keymap keymap)
   (set-extent-property extent 'highlight t)
-  (set-extent-property
-   extent 'help-echo
+  (set-extent-property extent 'dired-permissions (string-to-char (keymap-name keymap)))
+  (set-extent-property extent 'help-echo
    "button2 toggles, button3 changes otherwise."))
 
 (dired-setup-chmod-keymaps)
-	  
+
+
+;;; Popup header supports
+(defun dired-insert-set-headerline-properties ()
+  ;; Sets the extents for the current directory header line
+  (let ((position (dired-get-headerline)))
+    (dired-insert-set-headerline-properties-internal (car position)
+						     (cdr position))))
+  
+(defun dired-get-headerline ()
+  ;; Get the bufffer position of the current directory header line
+  (save-excursion
+    (beginning-of-line)
+    (save-match-data
+      (re-search-forward " *\\(.*\\):$" nil t)
+      ;; Return begin and end pair
+      (cons (match-beginning 1) (match-end 1)))))
+
+(defun dired-insert-set-headerline-properties-internal (b e)
+  (let ((extent (make-extent b e)))
+    (set-extent-face extent 'dired-face-header)
+    (set-extent-property extent 'keymap dired-headerline-local-map)
+    (set-extent-property extent 'highlight t)
+    (set-extent-property
+     extent 'help-echo
+     "button2 do dired this window.")))
+
+(defvar dired-headerline-local-map 
+  (let ((map (make-sparse-keymap)))
+    (set-keymap-name map 'dired-headerline-local-map)
+    (define-key map [button2] 'dired-headerline-dired)
+    (define-key map [(shift button2)] 'dired-headerline-dired-other-window)
+    (define-key map [(meta button2)] 'dired-headerline-dired-other-frame)
+    ;; 
+    (define-key map "f" 'dired-headerline-dired-by-key)
+    (define-key map "o" 'dired-headerline-dired-other-window-by-key)
+    (define-key map "w" 'dired-headerline-dired-frame-window-by-key)
+    map)
+  "Keymap used to activate actions on the header line in dired directory.")
+
+(defun dired-split-headerline (dir)
+  ;; (dired-split-headerline "/")
+  ;; => (list "/")
+  ;;
+  ;; (dired-split-headerline "/a/b/c")
+  ;; => (list  "/" "/a" "/a/b" "/a/b/c")
+  ;;
+  ;; (dired-split-headerline "/masat-y@foo:/a")
+  ;; => (list  "/"  "/masat-y@foo:/" "/masat-y@foo:/a")
+  ;; 
+  ;; (dired-split-headerline "/masat-y@foo:/")
+  ;; => (list  "/"  "/masat-y@foo:/")
+  
+  (let ((dirlist (cons dir nil))
+	(lastdir dir))			; To check double entry item
+    (setq dir (directory-file-name
+	       (file-name-directory  dir)))
+    (while (not (string= dir lastdir))
+      (setq dirlist (cons dir dirlist)
+	    lastdir dir
+	    dir (directory-file-name 
+		     (file-name-directory dir))))
+    ;; Add "/" if it is not in `dirlist'.
+    (if (not (string= (car dirlist) "/")) 
+	(setq dirlist (cons "/" dirlist)))
+    dirlist))
+
+;; by mouse
+(defun dired-headerline-dired (event)
+  "On the header line in dired, visit the parent directory."
+  (interactive "e")
+  (dired-headerline-dired-internal event 'dired "Dired"))
+
+(defun dired-headerline-dired-other-window (event)
+  "On the header line in dired, visit the parent directory in another window."
+  (interactive "e")
+  (dired-headerline-dired-internal event 
+				   'dired-other-window
+				   "Dired in Other Window"))
+
+(defun dired-headerline-dired-other-frame (event)
+  "On the header line in dired, visit the parent directory in another frame."
+  (interactive "e")
+  (dired-headerline-dired-internal event 
+				   'dired-other-frame
+				   "Dired in Other Frame"))
+
+(defun dired-headerline-dired-internal (event func title)
+  ;;
+  ;; Header line:  /a/b/c
+  ;; =>
+  ;; Popup menu: TITLE
+  ;;             ------
+  ;;             /a/b/c
+  ;;             /a/b
+  ;;             /a
+  ;;             /
+  ;;
+  (interactive "e")
+  (let (dirlist)
+    ;;
+    ;; Get directory list
+    ;;
+    (let (points basedir)
+      (save-excursion
+	(set-buffer (window-buffer (event-window event)))
+	(save-excursion
+	  (goto-char (event-point event))
+	  (setq points (dired-get-headerline)
+		dirlist (dired-split-headerline
+			 (buffer-substring (car points) (cdr points)))))))
+    ;;
+    ;; Create menu and popup it
+    ;;
+    (let (v items dir)
+      (while dirlist
+	(setq dir (car dirlist)
+	      dirlist (cdr dirlist)
+	      v (make-vector 3 0))
+	(aset v 0 dir) (aset v 1 (list func dir)) (aset v 2 t)
+	(setq items (cons v items)))
+      (popup-menu (cons title items)))))
+
+;; By key
+(defun dired-headerline-dired-by-key (p)
+  "On the header line in dired, visit the parent directory under ther cursor."
+  (interactive "d")
+  (dired-headerline-dired-by-key-internal p 'dired))
+
+(defun dired-headerline-dired-other-window-by-key (p)
+  "On the header line in dired, visit the parent directory under ther cursor on another window."
+  (interactive "d")
+  (dired-headerline-dired-by-key-internal p 'dired-other-window))
+
+(defun dired-headerline-dired-other-frame-by-key (p)
+  "On the header line in dired, visit the parent directory under ther cursor on another frame."
+  (interactive "d")
+  (dired-headerline-dired-by-key-internal p 'dired-other-frame))
+
+(defun dired-headerline-dired-by-key-internal (p func)
+  (let ((points (dired-get-headerline))
+	dirlist
+	basedir
+	dir
+	(check-remote nil))
+    ;; A function that checks the headerline shows remote directry.
+    (fset 'check-remote (lambda (str)
+			  (save-match-data (string-match "/.*@.*:/" str))))
+
+    ;; Calculate position in headerline
+    (setq p       (- p (car points)))
+
+    ;; Get directory list
+    (setq basedir (buffer-substring (car points) (cdr points)))
+    (setq dirlist (dired-split-headerline basedir))
+
+    ;; Which directory 'p' in on...
+    (setq dir
+	  (catch 'found
+	    ;;
+	    ;; P points out Root directory?
+	    (if (= p 0)
+		;; p is on "/"
+		(throw 'found "/") 
+	      ;; p is not on "/".
+	      (setq dirlist (cdr dirlist)))
+	    ;;
+	    ;; P points out remote host name field.
+	    ;;
+	    (if (and dirlist 
+		     (apply 'check-remote (list (car dirlist)))) 
+		;; P points out remote host field in headerline?
+		(if (< p (length (car dirlist))) 
+		    (throw 'found (car dirlist))
+		  ;; P points out remote directory not remote host field.
+		  (setq dirlist (cdr dirlist))))
+	    ;;
+	    ;; Find p  from directory list
+	    ;;
+	    (let (tmpdir)
+	      (while  dirlist
+		(setq tmpdir (car dirlist))
+		(if (<= p (length tmpdir))
+		    (throw 'found tmpdir)
+		  (setq dirlist (cdr dirlist))))
+	      nil)
+	    ))
+    ;;
+    ;; Run dired
+    ;;
+    (if (string= dir basedir)
+	(revert-buffer dir)
+      (funcall func dir))))
+
 ;;; end of dired-xemacs.el
